{
    +------------------------------------------------+
    |Filename: SCD30.spin                            |
    |Author: Phil Pilgrim                            |
    |Description: Driver for SCD30 CO2 Sensor        |
    |Copyright (c) 2022                              |
    |See end of file for terms of use.               |
    +------------------------------------------------+

  The SCD30 driver handles most functions of the Sensirion
  SCD30 CO2 sensor. It is designed to be mostly compatible with
  its sister driver for the SCD41. However, there are differences,
  that prevent it from being a drop-in replacement, particularly as they
  pertain to persistent settings. The user is urged to read the SCD30
  datasheet for complete details.

  This driver assumes that SCL, along with SDA, has a
  pull-up to Vdd. The i2c_phipi driver accompanying it accommodates
  clock-stretching from the slave device.

  2022.04.12 Initial beta release.

}

CON

  SCD30_CRC8_POLYNOMIAL = $131
  SCD30_CRC8_INIT       = $ff

  RECAL_SIZE            = 40
  I2C_ADDR              = $c2

VAR

  byte buffer[20]
  word ppms[RECAL_SIZE], temps[RECAL_SIZE], rhs[RECAL_SIZE], hi_ppm, lo_ppm, hi_temp, lo_temp, hi_rh, lo_rh

OBJ

  i2c:  "i2c_phipi"

PUB start(scl, sda)

  {{ Start routine. Sets pin numbers for scl and sda.
     Call first before doing anything else.
  }}

  i2c.init(scl, sda)

PUB configure

  {{ Default configuration. Continuous readings every 2 seconds.
     Writes to EEPROM and retains settings through power cycles,
     so use once only.
  }}

  set_interval(2)
  start_measurements

PUB start_measurements

  {{ Start periodic measurements every five seconds.
     Writes to EEPROM and retains settings through power cycles,
     so use once only.
  }}

  set_millibars(0)

PUB stop_measurements | data

  {{ Halts continuous measurements.
  }}

  send_cmd($0104)

PUB recalibrate(new_ppm) | i, ppm, temp, rh

  {{ Perform device recalibration to value specified by new_ppm.
     It waits for readings to equilibrate before sending the set_ppm command.
  }}

  repeat 120
    waitcnt(cnt + clkfreq)  
  repeat i from 0 to RECAL_SIZE - 1
    read_measurements
    ppms[i] := co2_ppm
    temps[i] := temp_c
    rhs[i] := rh_pct
  find_minmax
  repeat 500
    if (hi_ppm - lo_ppm < 5 and hi_temp - lo_temp < 10 and hi_rh - lo_rh < 3)
      set_ppm(new_ppm)
      return true
    read_measurements
    wordmove(@ppms[1], @ppms, RECAL_SIZE - 1)
    ppms := co2_ppm
    wordmove(@temps[1], @temps, RECAL_SIZE - 1)
    temps := temp_c
    wordmove(@rhs[1], @rhs, RECAL_SIZE - 1)
    rhs := rh_pct
    find_minmax
  return false

PUB set_altitude(alt)

  {{ Set the sensor's altitude in meters.
     This setting persists across power cycles.
     It writes to EEPROM, so use sparingly.
  }}

  send_cmd_data($5102, alt)

PUB get_altitude

  {{ Retrieve the current saved altitude.
  }}

  send_cmd($5102)
  receive_data(3)
  return buffer << 8 | buffer[1]

PUB set_millibars(mbars)

  {{ Set current air pressure in millibars.
  }} 

  send_cmd_data($0010, mbars)

PUB set_ppm(ppm)

  {{ Calibrate to a new PPM value.
     Be sure measurements have been running for at least 3 minutes before calling.
     Return true on success; false on failure.
  }}

  send_cmd_data($5204, ppm)

PUB set_temp_offset(offset)

  {{ Set the temperature offset in degrees C * 10.
     This setting persists across power cycles.
     It writes to EEPROM, so use sparingly.
  }}

  send_cmd_data($5403, offset)

PUB set_interval(time)

  {{ Set the interval between measurements in seconds (2 - 1800).
  }}

  send_cmd_data($4600, time)

PUB self_calibrate(yesno)

  {{ Set or clear self-calibration.
     Setting persists trhough power cycles.
     It writes to EEPROM, so use sparingly. 
  }}

  send_cmd_data($5306, yesno)

PUB data_ready

  {{ Check to see if new data are ready.
     Returns TRUE if so; FALSE if not.
  }}

  send_cmd($0202)
  receive_data(3)
  return buffer[1] > 0

PUB read_measurements 

  {{ Read ppm, humidity, and temp from sensor.
  }}

  repeat until data_ready
  send_cmd($0300)
  receive_data(18)

PUB co2_ppm

  {{ Return CO2 concentration in PPM.
  }}

  return flt2int(buffer[0] << 24 | buffer[1] << 16 | buffer[3] << 8 | buffer[4], 1)

PUB temp_f

  {{ Return temperature in F * 10.
  }}

  return temp_c * 9 / 5 + 320

PUB temp_c

  {{ Return temperature in C * 10.
  }}

  return flt2int(buffer[6] << 24 | buffer[7] << 16 | buffer[9] << 8 | buffer[10], 10)

PUB rh_pct

  {{ Return relative humidity in whole percent.
  }} 

  return flt2int(buffer[12] << 24 | buffer[13] << 16 | buffer[15] << 8 | buffer[16], 1)

PUB firmware_version

  {{ Return SCD30's firmware version.
  }}
  
  send_cmd($d100)
  receive_data(3)
  return buffer << 8 | buffer[1]

PRI find_minmax | i

  hi_ppm := hi_temp := hi_rh := 0
  lo_ppm := lo_temp := lo_rh := 65535
  repeat i from 0 to RECAL_SIZE - 1
    hi_ppm #>= ppms[i]
    lo_ppm <#= ppms[i]
    hi_temp #>= temps[i]
    lo_temp <#= temps[i]
    hi_rh #>= rhs[i]
    lo_rh <#= rhs[i]
 
PRI send_cmd(cmd) | data

  i2c.send_addr(I2C_ADDR)
  byte[@data] := cmd >> 8
  byte[@data + 1] := cmd
  i2c.send_bytes(@data, 2)
  i2c.stop

PRI send_cmd_data(cmd, value) | data

  i2c.send_addr(I2C_ADDR)
  byte[@data] := cmd >> 8
  byte[@data + 1] := cmd
  i2c.send_bytes(@data, 2)
  byte[@data] := value >> 8
  byte[@data + 1] := value
  byte[@data + 2] := find_crc(@data, 2)
  i2c.send_bytes(@data, 3)
  i2c.stop

PRI receive_data(nbytes)

  wait_ms(5)
  i2c.send_addr(I2C_ADDR | 1)
  i2c.receive_bytes(@buffer, nbytes, i2c#NAK)
  i2c.stop
   
PRI find_crc(array_addr, nbytes) : crc

  crc := SCD30_CRC8_INIT
  repeat nbytes
    crc ^= byte[array_addr++]
    repeat 8
      crc <<= 1
      if (crc & $100)
        crc ^= SCD30_CRC8_POLYNOMIAL
  crc &= $ff

PRI wait_ms(milliseconds)

  waitcnt(cnt + clkfreq / 1000 * milliseconds)

PRI flt2int(flt, scale) | sign, exp

  if (flt == 0.0)
    return 0
  sign := flt >> 31
  exp := flt >> 23 & $ff - 149
  result := (((flt & $7fffff | $800000) * scale) + 1) >> 1
  if (exp < 0)
    result >>= -exp
  elseif (exp > 0)
    result <<= exp
  if (sign)
    -result
{
    --------------------------------------------------------------------------------------------------------
    TERMS OF USE: MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
    associated documentation files (the "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the
    following conditions:

    The above copyright notice and this permission notice shall be included in all copies or substantial
    portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
    LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
    --------------------------------------------------------------------------------------------------------
}   